package auth

import (
	"crypto/sha256"
	"encoding/base64"
	"errors"

	"golang.org/x/crypto/ssh"

	"git.tcp.direct/kayos/sh3lly/internal/sanitize"
)

func (users *UserDB) keyboardInteractive(conn ssh.ConnMetadata, challenge ssh.KeyboardInteractiveChallenge) (*ssh.Permissions, error) {
	client := sanitize.Data(string(conn.ClientVersion()), 64)
	slog := log.With().
		Str("caller", conn.RemoteAddr().String()).
		Str("client", client).Str("user", conn.User()).Logger()

	var err error
	if users.AcceptPassphrase() {
		var answers []string
		answers, err = challenge("", "", []string{"Password: "}, []bool{false})
		if err == nil {
			if len(answers) != 1 {
				err = errors.New("missing input")
			} else {
				err = users.PassphraseLogin(conn.User(), answers[0])
				if err != nil {
					slog.Warn().Err(err).Msg("password auth failure")
					return nil, err
				}
			}
		}
	}
	if !users.AllowAnonymous() {
		err = errors.New("public key authentication required")
	}

	err = users.CheckBans(conn.User(), conn.RemoteAddr(), nil, client)
	if err != nil {
		slog.Warn().Msg(err.Error())
		return nil, err
	}

	slog.Debug().Msg("not banned")

	return nil, err
}

func (users *UserDB) publicKeyCallback(conn ssh.ConnMetadata, key ssh.PublicKey) (*ssh.Permissions, error) {
	err := users.CheckBans(sanitize.Data(conn.User(), 32), conn.RemoteAddr(), key, sanitize.Data(string(conn.ClientVersion()), 64))
	if err != nil {
		return nil, err
	}
	err = users.CheckPublicKey(conn.User(), key)
	if err != nil {
		return nil, err
	}
	perm := &ssh.Permissions{Extensions: map[string]string{
		"pubkey": string(key.Marshal()),
	}}
	return perm, nil
}

// MakeAuth makes an ssh.ServerConfig which performs authentication against an Auth implementation.
// TODO: Switch to using ssh.AuthMethod instead?
func MakeAuth(auth Authenticator) *ssh.ServerConfig {
	config := ssh.ServerConfig{
		NoClientAuth: false,

		// Auth-related things should be constant-time to avoid timing attacks.
		PublicKeyCallback: auth.publicKeyCallback,

		// We use KeyboardInteractiveCallback instead of PasswordCallback to
		// avoid preventing the client from including a pubkey in the user
		// identification.
		KeyboardInteractiveCallback: auth.keyboardInteractive,
	}

	return &config
}

// MakeNoAuth makes a simple ssh.ServerConfig which allows all connections.
// Primarily used for testing.
func MakeNoAuth() *ssh.ServerConfig {
	config := ssh.ServerConfig{
		NoClientAuth: false,
		// Auth-related things should be constant-time to avoid timing attacks.
		PublicKeyCallback: func(conn ssh.ConnMetadata, key ssh.PublicKey) (*ssh.Permissions, error) {
			perm := &ssh.Permissions{Extensions: map[string]string{
				"pubkey": string(key.Marshal()),
			}}
			return perm, nil
		},
		KeyboardInteractiveCallback: func(conn ssh.ConnMetadata, challenge ssh.KeyboardInteractiveChallenge) (*ssh.Permissions, error) {
			return nil, nil
		},
	}

	return &config
}

// Fingerprint performs a SHA256 BASE64 fingerprint of the PublicKey, similar to OpenSSH.
// See: https://anongit.mindrot.org/openssh.git/commit/?id=56d1c83cdd1ac
func Fingerprint(k ssh.PublicKey) string {
	hash := sha256.Sum256(k.Marshal())
	return "SHA256:" + base64.StdEncoding.EncodeToString(hash[:])
}
