package config

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"runtime"

	"github.com/rs/zerolog"
	"github.com/spf13/viper"
)


func init() {
	prefConfigLocation = "./"
	snek = viper.New()
}

func writeConfig() {
	if runtime.GOOS != "windows" {
		if _, err := os.Stat(prefConfigLocation); os.IsNotExist(err) {
			if err = os.Mkdir(prefConfigLocation, 0755); err != nil {
				println("error writing new config: " + err.Error())
			}
		}
		newconfig := prefConfigLocation + "/" + "config.toml"
		if err = snek.SafeWriteConfigAs(newconfig); err != nil {
			fmt.Println(err.Error())
			os.Exit(1)
		}
		Filename = newconfig
		return
	}

	newconfig := Title + "-config"
	snek.SetConfigName(newconfig)
	if err = snek.MergeInConfig(); err != nil {
		if err = snek.SafeWriteConfigAs(newconfig + ".toml"); err != nil {
			fmt.Println(err.Error())
			os.Exit(1)
		}
	}

	Filename = newconfig
}

// Init will initialize our toml configuration engine and define our default configuration values which can be written to a new configuration file if desired
func Init() {
	snek.SetConfigType("toml")
	snek.SetConfigName("config")

	argParse()

	if customconfig {
		associateExportedVariables()
		return
	}

	setConfigFileLocations()
	setDefaults()

	for _, loc := range configLocations {
		snek.AddConfigPath(loc)
	}

	if err = snek.MergeInConfig(); err != nil {
		writeConfig()
	}

	if len(Filename) < 1 {
		Filename = snek.ConfigFileUsed()
	}

	associateExportedVariables()
}

func setDefaults() {
	var (
		configSections = []string{"logger", "ssh", "database", "shells"}
		deflogdir      = "./logs/"
		defNoColor     = false
	)

	if runtime.GOOS == "windows" {
		deflogdir = "logs/"
		defNoColor = true
	}

	Opt := make(map[string]map[string]interface{})

	Opt["logger"] = map[string]interface{}{
		"debug":             true,
		"directory":         deflogdir,
		"nocolor":           defNoColor,
		"use_date_filename": true,
	}

	Opt["ssh"] = map[string]interface{}{
		"bind_addr":       "127.0.0.1",
		"bind_port":       2222,
		"key_path":        "./",
		"allow_anonymous": true,
		"server_version":  "tcp.direct",
	}

	Opt["shells"] = map[string]interface{}{
		"bind_addr": "192.168.69.5",
		"bind_port": 4444,
	}

	Opt["database"] = map[string]interface{}{
		"datapath": "./data",
	}

	for _, def := range configSections {
		snek.SetDefault(def, Opt[def])
	}

	if GenConfig {
		if err = snek.SafeWriteConfigAs("./config.toml"); err != nil {
			fmt.Println(err.Error())
			os.Exit(1)
		}
		os.Exit(0)
	}

}

func setConfigFileLocations() {
	configLocations = append(configLocations, "./")

	if runtime.GOOS != "windows" {
		configLocations = append(configLocations, prefConfigLocation)
		configLocations = append(configLocations, "/etc/"+Title+"/")
		configLocations = append(configLocations, "../")
		configLocations = append(configLocations, "../../")
	}
}

func loadCustomConfig(path string) {
	if f, err = os.Open(path); err != nil {
		println("Error opening specified config file: " + path)
		panic("config file open fatal error: " + err.Error())
	}
	buf, err := ioutil.ReadAll(f)
	err2 := snek.ReadConfig(bytes.NewBuffer(buf))
	switch {
	case err != nil:
		fmt.Println("config file read fatal error: ", err.Error())
	case err2 != nil:
		fmt.Println("config file read fatal error: ", err2.Error())
	default:
		break
	}
	customconfig = true
}

func printUsage() {
	println("\n" + Title + " v" + Version + " Usage\n")
	println("-c <config.toml> - Specify config file")
	println("--nocolor - disable color and banner ")
	println("--banner - show banner + version and exit")
	println("--genconfig - write default config to 'default.toml' then exit")
	os.Exit(0)
}

// TODO: should probably just make a proper CLI with flags or something
func argParse() {
	for i, arg := range os.Args {
		switch arg {
		case "-h":
			printUsage()
		case "--genconfig":
			GenConfig = true
		case "--nocolor":
			noColorForce = true
		case "--config":
			fallthrough
		case "-c":
			if len(os.Args) <= i-1 {
				panic("syntax error! expected file after -c")
			}
			loadCustomConfig(os.Args[i+1])
		default:
			continue
		}
	}
}

func processOpts() {
	// string options and their exported variables
	stringOpt := map[string]*string{
		"ssh.bind_addr":      &BindAddr,
		"ssh.key_path":       &KeyPath,
		"ssh.server_version": &SSHVersion,
		"shells.bind_addr":   &ShellAddr,
		"logger.directory":   &LogDir,
		"database.datapath":  &DataPath,
	}

	// bool options and their exported variables
	boolOpt := map[string]*bool{
		"logger.debug":        &Debug,
		"logger.nocolor":      &NoColor,
		"ssh.allow_anonymous": &AllowAnon,
	}

	intOpt := map[string]*int{
		"ssh.bind_port":    &BindPort,
		"shells.bind_port": &ShellPort,
	}

	for key, opt := range stringOpt {
		*opt = snek.GetString(key)
	}
	for key, opt := range boolOpt {
		*opt = snek.GetBool(key)
	}
	/*	for key, opt := range strSliceOpt {
			*opt = snek.GetStringSlice(key)
	}*/
	for key, opt := range intOpt {
		*opt = snek.GetInt(key)
	}
}

func associateExportedVariables() {
	processOpts()

	if noColorForce {
		NoColor = true
	}

	if Debug {
		zerolog.SetGlobalLevel(zerolog.DebugLevel)
	}
}
